/*
 * Copyright (c) 2010-2024. Institut Pasteur.
 *
 * This file is part of Icy.
 * Icy is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Icy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Icy. If not, see <https://www.gnu.org/licenses/>.
 */

package plugins.adufour.blocks.tools.io;

import org.apache.poi.ss.usermodel.*;
import org.bioimageanalysis.icy.extension.plugin.abstract_.Plugin;
import org.bioimageanalysis.icy.extension.plugin.annotation_.IcyPluginIcon;
import org.bioimageanalysis.icy.extension.plugin.annotation_.IcyPluginName;
import org.jetbrains.annotations.NotNull;
import plugins.adufour.blocks.tools.io.WorkbookToFile.MergePolicy;
import plugins.adufour.blocks.util.VarList;
import plugins.adufour.vars.lang.VarFile;
import plugins.adufour.vars.lang.VarWorkbook;
import plugins.adufour.workbooks.Workbooks;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;

/**
 * IO block that reads a workbook file from disk using the POI library
 *
 * @author Alexandre Dufour
 */
@IcyPluginName("File to Workbook")
@IcyPluginIcon(path = "/workbooks.png")
public class FileToWorkbook extends Plugin implements IOBlock {
    VarFile inputFile = new VarFile("input file", null);
    VarWorkbook workbook = new VarWorkbook("workbook", (Workbook) null);

    @Override
    public void run() {
        workbook.setValue(readWorkbook(inputFile.getValue(true)));
    }

    @Override
    public void declareInput(final @NotNull VarList inputMap) {
        inputMap.add("input file", inputFile);
    }

    @Override
    public void declareOutput(final @NotNull VarList outputMap) {
        outputMap.add("workbook", workbook);
    }

    /**
     * Reads the specified file into a workbook. This method will read Excel-compatible (.xls)
     * files, but will also read tab-delimited text files. <br>
     * Note: tab-delimited text files may contain multiple sheets, if specified as in the following
     * example:<br>
     * <code>== sheet 1 ==<br>
     * item1 (tab) item2<br>
     * == sheet 2 ==<br>
     * item3 (tab) item4</code><br>
     * (the same convention is followed by
     * {@link WorkbookToFile#saveAsText(Workbook, String, MergePolicy)}
     *
     * @param file input file
     * @return the workbook read from input file
     */
    public static Workbook readWorkbook(final @NotNull File file) {
        if (!file.exists() || file.isDirectory())
            throw new IllegalArgumentException("Cannot read a workbook from " + file.getPath());

        try {
            final FileInputStream fis = new FileInputStream(file);
            final Workbook workbook = WorkbookFactory.create(fis);
            fis.close();
            return workbook;
        }
        catch (final Exception e) {
            // try text
            final Workbook wb = Workbooks.createEmptyWorkbook();
            Sheet sheet = null;

            try {
                final BufferedReader reader = new BufferedReader(new FileReader(file));
                int rowID = 0;
                boolean newLine = false;

                while (reader.ready()) {
                    final String line = reader.readLine();

                    if (line.isEmpty()) {
                        // remember there is a new line, deal with it later
                        newLine = true;
                        continue;
                    }

                    // 1) fetch (or create) the sheet

                    // special case: sheet name saved using FileToWorkbook
                    if (line.startsWith("== ")) {
                        // don't add the new line if we create a new sheet
                        newLine = false;
                        final String sheetName = line.substring(3, line.indexOf(" =="));
                        sheet = wb.createSheet(sheetName);

                        // reset the row
                        rowID = 0;

                        // nothing else interesting in this line...
                        continue;
                    }

                    // otherwise, create a (default) empty sheet
                    if (sheet == null) {
                        sheet = wb.createSheet();
                        // reset the row
                        rowID = 0;
                    }

                    // 2) add the new line (if any)
                    if (newLine)
                        sheet.createRow(rowID++);

                    // 3) Create a new row for the current input
                    final Row row = sheet.createRow(rowID++);

                    // skip empty lines
                    if (line.isEmpty())
                        continue;

                    int colID = 0;

                    final String[] words = line.split("\t");
                    for (final String word : words) {
                        if (!word.isEmpty()) {
                            final Cell cell = row.createCell(colID);
                            try {
                                cell.setCellValue(Double.parseDouble(word));
                            }
                            catch (final NumberFormatException formatError) {
                                cell.setCellValue(word);
                            }
                        }
                        colID++;
                    }
                }

                reader.close();

                return wb;
            }
            catch (final Exception e2) {
                throw new RuntimeException(e2);
            }
        }
    }
}
