/*
 * Copyright (c) 2010-2024. Institut Pasteur.
 *
 * This file is part of Icy.
 * Icy is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Icy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Icy. If not, see <https://www.gnu.org/licenses/>.
 */

package plugins.adufour.vars.gui.swing;

import org.apache.poi.ss.formula.FormulaParseException;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.ss.usermodel.WorkbookFactory;
import org.bioimageanalysis.icy.common.math.ArrayMath;
import org.bioimageanalysis.icy.gui.component.button.IcyButton;
import org.bioimageanalysis.icy.gui.component.icon.SVGIcon;
import org.bioimageanalysis.icy.gui.dialog.MessageDialog;
import org.bioimageanalysis.icy.gui.dialog.OpenDialog;
import org.bioimageanalysis.icy.gui.frame.IcyFrame;
import org.bioimageanalysis.icy.system.IcyHandledException;
import org.bioimageanalysis.icy.system.thread.ThreadUtil;
import org.jdesktop.swingx.JXPanel;
import org.jdesktop.swingx.JXTable;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.math.plot.Plot2DPanel;
import org.math.plot.Plot3DPanel;
import org.math.plot.utils.Array;
import plugins.adufour.blocks.tools.io.WorkbookToFile;
import plugins.adufour.blocks.tools.io.WorkbookToFile.MergePolicy;
import plugins.adufour.ezplug.EzButton;
import plugins.adufour.ezplug.EzDialog;
import plugins.adufour.ezplug.EzLabel;
import plugins.adufour.ezplug.EzVarEnum;
import plugins.adufour.vars.gui.VarEditor;
import plugins.adufour.vars.lang.Var;
import plugins.adufour.vars.lang.VarBoolean;
import plugins.adufour.vars.lang.VarString;
import plugins.adufour.vars.lang.VarWorkbook;
import plugins.adufour.vars.util.VarListener;
import plugins.adufour.workbooks.IcySpreadSheet;
import plugins.adufour.workbooks.Workbooks;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.table.*;
import java.awt.*;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;

public class WorkbookEditor extends SwingVarEditor<Workbook> {
    final VarString formula = new VarString("Formula", "");
    final VarBoolean useHeader = new VarBoolean("Header", false);
    final VarBoolean readOnly = new VarBoolean("read-only", false);
    final VarBoolean openButtonVisible = new VarBoolean("show open button", true) {
        @Override
        public void valueChanged(final Var<Boolean> source, final Boolean oldValue, final Boolean newValue) {
            super.valueChanged(source, oldValue, newValue);
            if (openButton != null)
                ThreadUtil.invokeLater(() -> openButton.setVisible(false));
        }
    };

    private JButton openButton, exportButton, plotButton;
    private JTabbedPane tabs;
    private final HashMap<Sheet, JXTable> tables = new HashMap<>();
    private final HashMap<Sheet, JXTable> headers = new HashMap<>();

    // cached workbook to avoid reference of old book being leaked by GUI
    private Workbook book;

    class TabChangeListener implements ChangeListener {
        @Override
        public void stateChanged(final ChangeEvent arg0) {
            final Workbook book = variable.getValue();
            final int sheetIndex = tabs.getSelectedIndex();

            if (sheetIndex == book.getNumberOfSheets()) {
                tabs.removeTabAt(book.getNumberOfSheets());
                createSheet(book);
                updateInterfaceValue();
                tabs.setSelectedIndex(book.getNumberOfSheets() - 1);
                book.setActiveSheet(book.getNumberOfSheets() - 1);
            }

            if (sheetIndex >= 0) {
                final Sheet sheet = book.getSheetAt(sheetIndex);
                book.setActiveSheet(sheetIndex);
                if (tables.containsKey(sheet)) {
                    final JXTable table = tables.get(sheet);
                    final SheetModel model = (SheetModel) table.getModel();
                    updateFormulaField(model, table.getSelectedRow(), table.getSelectedColumn());
                }
            }
        }
    }

    private final TabChangeListener tabChangeListener = new TabChangeListener();

    public WorkbookEditor(final VarWorkbook variable) {
        super(variable);
    }

    @Override
    protected JComponent createEditorComponent() {
        return new JXPanel(new BorderLayout());
    }

    @Override
    protected void activateListeners() {
        useHeader.addListener(new VarListener<>() {
            @Override
            public void valueChanged(final Var<Boolean> source, final Boolean oldValue, final Boolean newValue) {
                updateInterfaceValue();
            }

            @Override
            public void referenceChanged(final Var<Boolean> source, final Var<? extends Boolean> oldReference, final Var<? extends Boolean> newReference) {
            }
        });
    }

    @Override
    protected void deactivateListeners() {
        useHeader.removeListeners();
    }

    @Override
    public void dispose() {
        super.dispose();

        tables.clear();
        headers.clear();
        book = null;

        if (tabs != null) {
            tabs.removeChangeListener(tabChangeListener);
            tabs.removeAll();
        }
    }

    @Override
    public Dimension getPreferredSize() {
        return new Dimension(400, 100);
    }

    @Override
    public double getComponentVerticalResizeFactor() {
        return 1.0;
    }

    @Override
    public boolean isComponentResizeable() {
        return true;
    }

    /**
     * Indicates whether the first table row should be used as a header. If so, the first row will
     * virtually disappear from the table and its values will be used to rename the table columns
     *
     * @param header <code>true</code> to use the first row as a header, <code>false</code> otherwise
     */
    public void setFirstRowAsHeader(final boolean header) {
        useHeader.setValue(header);
    }

    /**
     * Sets whether the open button should be visible
     *
     * @param visible visible state
     */
    public void setOpenButtonVisible(final boolean visible) {
        openButtonVisible.setValue(visible);
    }

    /**
     * Sets whether the table editor is read-only, and therefore rejects user input
     *
     * @param readOnly read only state
     */
    public void setReadOnly(final boolean readOnly) {
        this.readOnly.setValue(readOnly);
    }

    private @Nullable MergePolicy confirm() {
        final EzDialog confirmDialog = new EzDialog("Confirmation");

        confirmDialog.addEzComponent(new EzLabel("This file already exists."));
        final EzVarEnum<MergePolicy> policy = new EzVarEnum<>("Action: ", MergePolicy.values());
        confirmDialog.addEzComponent(policy);
        final VarBoolean confirmed = new VarBoolean("confirmed", true);
        confirmDialog.addEzComponent(new EzButton("Ok", e -> {
            confirmed.setValue(true);
            confirmDialog.close();
        }));
        confirmDialog.addEzComponent(new EzButton("Cancel", e -> {
            confirmed.setValue(false);
            confirmDialog.close();
        }));

        // show confirm dialog and wait for user choice...
        confirmDialog.showDialog(true);
        confirmDialog.hideDialog();

        if (confirmed.getValue())
            return policy.getValue();

        return null;
    }

    @Override
    protected void updateInterfaceValue() {
        if (variable.getReference() != null) {
            setReadOnly(true);
            setOpenButtonVisible(false);
        }

        // update cached book
        book = variable.getValue();

        if (book == null)
            return;

        final JXPanel wbPanel = (JXPanel) getEditorComponent();
        final String bookID = "" + book.hashCode();

        final boolean buildNewEditor = (wbPanel.getName() == null || !wbPanel.getName().equalsIgnoreCase(bookID));

        if (buildNewEditor) {
            // create a new table
            wbPanel.removeAll();

            // Option panel
            final JPanel optionBar = new JPanel();
            optionBar.setLayout(new BoxLayout(optionBar, BoxLayout.X_AXIS));

            // Open
            openButton = new IcyButton(SVGIcon.FILE_OPEN);
            openButton.setVisible(openButtonVisible.getValue());
            openButton.setBorderPainted(false);
            openButton.setFocusable(false);
            openButton.setContentAreaFilled(false);
            openButton.setToolTipText("Open/Create workbook...");
            openButton.addActionListener(e -> {
                final String path = OpenDialog.chooseFile("Load/Create workbook...", null, "Workbook", ".xls");
                if (path == null)
                    return;

                try {
                    variable.setValue(WorkbookFactory.create(new File(path)));
                }
                catch (final Exception e1) {
                    MessageDialog.showDialog(e1.getMessage(), MessageDialog.ERROR_MESSAGE);
                }
            });
            optionBar.add(openButton);

            // Export
            exportButton = new IcyButton(SVGIcon.SAVE);
            exportButton.setBorderPainted(false);
            exportButton.setFocusable(false);
            exportButton.setContentAreaFilled(false);
            exportButton.setToolTipText("Export workbook to disk...");

            final JFileChooser fileChooser = new JFileChooser();
            final FileNameExtensionFilter txtFilter = new FileNameExtensionFilter("Text files (.txt)", "txt");
            final FileNameExtensionFilter xlsFilter = new FileNameExtensionFilter("Spreadsheets (.xlsx)", "xlsx");

            final JPopupMenu exportPopupMenu = new JPopupMenu();
            final JMenuItem exportXLS = new JMenuItem("Export to an Excel file");
            exportXLS.addActionListener(e -> {
                fileChooser.setFileFilter(xlsFilter);
                if (fileChooser.showSaveDialog(exportPopupMenu) != JFileChooser.APPROVE_OPTION)
                    return;
                final File outputFile = fileChooser.getSelectedFile();
                if (outputFile.exists()) {
                    // ask for confirmation + merging policy
                    final MergePolicy mergingPolicy = confirm();
                    // confirmed ?
                    if (mergingPolicy != null)
                        WorkbookToFile.saveAsSpreadSheet(book, outputFile.getPath(), mergingPolicy);
                }
                else
                    WorkbookToFile.saveAsSpreadSheet(book, outputFile.getPath(), MergePolicy.Overwrite);
            });
            final JMenuItem exportTXT = new JMenuItem("Export to a text file");
            exportTXT.addActionListener(e -> {
                fileChooser.setFileFilter(txtFilter);
                if (fileChooser.showSaveDialog(exportPopupMenu) != JFileChooser.APPROVE_OPTION)
                    return;
                final File outputFile = fileChooser.getSelectedFile();
                if (outputFile.exists()) {
                    // ask for confirmation + merging policy
                    final MergePolicy mergingPolicy = confirm();
                    // confirmed ?
                    if (mergingPolicy != null)
                        WorkbookToFile.saveAsText(book, outputFile.getPath(), mergingPolicy);
                }
                else
                    WorkbookToFile.saveAsText(book, outputFile.getPath(), MergePolicy.Overwrite);
            });
            exportPopupMenu.add(exportTXT);
            exportPopupMenu.add(exportXLS);
            exportButton.addActionListener(e -> exportPopupMenu.show(exportButton, 0, exportButton.getHeight()));
            optionBar.add(exportButton);

            // Plot (START)

            plotButton = new IcyButton(SVGIcon.FORMAT_ALIGN_LEFT);
            plotButton.setBorderPainted(false);
            plotButton.setFocusable(false);
            plotButton.setContentAreaFilled(false);
            plotButton.setToolTipText("Plot");

            final JPopupMenu plotPopupMenu = new JPopupMenu();

            final JMenu plotHistogram1D = new JMenu("Histogram (1D)");
            plotHistogram1D.addMenuListener(new MenuListener() {
                @Override
                public void menuSelected(final MenuEvent e) {
                    if (book.getNumberOfSheets() == 0)
                        return;

                    final int sheetIndex = book.getActiveSheetIndex();
                    if (sheetIndex == -1)
                        return;

                    final Sheet sheet = book.getSheetAt(sheetIndex);
                    final IcySpreadSheet icySheet = Workbooks.getSheet(book, book.getSheetName(sheetIndex));

                    for (int i = 0; i < icySheet.getNumberOfColumns(); i++) {
                        final double[] histData = icySheet.getColumnValues(i);
                        if (histData == null || histData.length < 2)
                            continue;

                        final String columnName = tables.get(sheet).getColumnName(i);
                        final JMenuItem plotColHistogram = new JMenuItem(columnName);
                        plotColHistogram.addActionListener(actionEvent -> {
                            final double min = ArrayMath.min(histData);
                            final double max = ArrayMath.max(histData);
                            final int bins = Math.max(10, (int) Math.round(Math.sqrt(histData.length)));
                            final Plot2DPanel panel = new Plot2DPanel();
                            panel.addHistogramPlot(columnName, histData, min, max, bins);
                            panel.setFixedBounds(0, min, max);

                            // HistogramChart histoChart = new HistogramChart(histData);
                            // Group group = new Group(histoChart);
                            // final Scene scene = new Scene(group);
                            // JFXPanel panel = new JFXPanel();
                            // panel.setLayout(new BorderLayout());
                            // panel.setScene(scene);

                            final IcyFrame plot = new IcyFrame("Histogram of " + columnName, true);
                            plot.setPreferredSize(new Dimension(520, 420));
                            plot.addToDesktopPane();
                            plot.add(panel);
                            plot.pack();
                            plot.setVisible(true);
                        });
                        plotHistogram1D.add(plotColHistogram);
                    }
                }

                @Override
                public void menuDeselected(final MenuEvent e) {
                    plotHistogram1D.removeAll();
                }

                @Override
                public void menuCanceled(final MenuEvent e) {
                }
            });

            final JMenu plotHistogram2D = new JMenu("Histogram (2D)");
            plotHistogram2D.addMenuListener(new MenuListener() {
                @Override
                public void menuSelected(final MenuEvent e) {
                    if (book.getNumberOfSheets() == 0)
                        return;

                    final int sheetIndex = book.getActiveSheetIndex();
                    if (sheetIndex == -1)
                        return;

                    final Sheet sheet = book.getSheetAt(sheetIndex);
                    final IcySpreadSheet icySheet = Workbooks.getSheet(book, book.getSheetName(sheetIndex));

                    final JMenuItem labelFirstColumn = new JMenuItem("Select X axis...");
                    labelFirstColumn.setEnabled(false);
                    plotHistogram2D.add(labelFirstColumn);
                    plotHistogram2D.addSeparator();
                    for (int i = 0; i < icySheet.getNumberOfColumns(); i++) {
                        final double[] xValues = icySheet.getColumnValues(i);
                        if (xValues == null)
                            continue;

                        final double minX = ArrayMath.min(xValues);
                        final double maxX = ArrayMath.max(xValues);
                        final int binsX = Math.max(10, (int) Math.round(Math.sqrt(xValues.length)));

                        final String xColumnName = tables.get(sheet).getColumnName(i);
                        final JMenu plotFirstColumn = new JMenu(xColumnName);
                        plotHistogram2D.add(plotFirstColumn);

                        final JMenuItem labelSecondColumn = new JMenuItem("Select Y axis...");
                        labelSecondColumn.setEnabled(false);
                        plotFirstColumn.add(labelSecondColumn);
                        plotFirstColumn.addSeparator();
                        for (int j = 0; j < icySheet.getNumberOfColumns(); j++) {
                            if (i == j)
                                continue;

                            final double[] yValues = icySheet.getColumnValues(j);
                            if (yValues == null)
                                continue;

                            if (xValues.length != yValues.length)
                                throw new IcyHandledException("Cannot create scatter plot: datasets have different sizes");

                            final double minY = ArrayMath.min(yValues);
                            final double maxY = ArrayMath.max(yValues);
                            final int binsY = Math.max(10, (int) Math.round(Math.sqrt(yValues.length)));

                            final String yColumnName = tables.get(sheet).getColumnName(j);
                            final String plotTitle = xColumnName + " vs. " + yColumnName;

                            final JMenuItem plotSecondColumn = new JMenuItem(yColumnName);
                            plotSecondColumn.addActionListener(actionEvent -> {
                                final Plot3DPanel panel = new Plot3DPanel();
                                final double[][] xy = Array.mergeColumns(xValues, yValues);
                                panel.addHistogramPlot(plotTitle, xy, binsX, binsY);
                                panel.setFixedBounds(new double[]{minX, minY}, new double[]{maxX, maxY});

                                final IcyFrame plot = new IcyFrame(plotTitle, true);
                                plot.setPreferredSize(new Dimension(520, 420));
                                plot.addToDesktopPane();
                                plot.add(panel);
                                plot.pack();
                                plot.setVisible(true);
                            });
                            plotFirstColumn.add(plotSecondColumn);
                        }
                    }
                }

                @Override
                public void menuDeselected(final MenuEvent e) {
                    plotHistogram2D.removeAll();
                }

                @Override
                public void menuCanceled(final MenuEvent e) {
                }
            });

            final JMenu plotScatter2D = new JMenu("Scatter plot (2D)");
            plotScatter2D.addMenuListener(new MenuListener() {
                @Override
                public void menuSelected(final MenuEvent e) {
                    if (book.getNumberOfSheets() == 0)
                        return;

                    final int sheetIndex = book.getActiveSheetIndex();
                    if (sheetIndex == -1)
                        return;

                    final Sheet sheet = book.getSheetAt(sheetIndex);
                    final IcySpreadSheet icySheet = Workbooks.getSheet(book, book.getSheetName(sheetIndex));

                    final JMenuItem labelFirstColumn = new JMenuItem("Select X axis...");
                    labelFirstColumn.setEnabled(false);
                    plotScatter2D.add(labelFirstColumn);
                    plotScatter2D.addSeparator();
                    for (int i = 0; i < icySheet.getNumberOfColumns(); i++) {
                        final double[] xValues = icySheet.getColumnValues(i);
                        if (xValues == null)
                            continue;

                        final double minX = ArrayMath.min(xValues);
                        final double maxX = ArrayMath.max(xValues);

                        final String xColumnName = tables.get(sheet).getColumnName(i);
                        final JMenu plotFirstColumn = new JMenu(xColumnName);
                        plotScatter2D.add(plotFirstColumn);

                        final JMenuItem labelSecondColumn = new JMenuItem("Select Y axis...");
                        labelSecondColumn.setEnabled(false);
                        plotFirstColumn.add(labelSecondColumn);
                        plotFirstColumn.addSeparator();
                        for (int j = 0; j < icySheet.getNumberOfColumns(); j++) {
                            if (i == j)
                                continue;

                            final double[] yValues = icySheet.getColumnValues(j);
                            if (yValues == null)
                                continue;

                            if (xValues.length != yValues.length)
                                throw new IcyHandledException("Cannot create scatter plot: datasets have different sizes");

                            final double minY = ArrayMath.min(yValues);
                            final double maxY = ArrayMath.max(yValues);

                            final String yColumnName = tables.get(sheet).getColumnName(j);
                            final JMenuItem plotSecondColumn = new JMenuItem(yColumnName);
                            plotSecondColumn.addActionListener(actionEvent -> {
                                final Plot2DPanel panel = new Plot2DPanel();
                                panel.addScatterPlot("Scatter plot", xValues, yValues);
                                panel.setAxisLabels(xColumnName, yColumnName);
                                panel.setFixedBounds(new double[]{minX, minY}, new double[]{maxX, maxY});

                                final IcyFrame plot = new IcyFrame(xColumnName + " vs. " + yColumnName, true);
                                plot.setPreferredSize(new Dimension(520, 420));
                                plot.addToDesktopPane();
                                plot.add(panel);
                                plot.pack();
                                plot.setVisible(true);
                            });
                            plotFirstColumn.add(plotSecondColumn);
                        }
                    }
                }

                @Override
                public void menuDeselected(final MenuEvent e) {
                    plotScatter2D.removeAll();
                }

                @Override
                public void menuCanceled(final MenuEvent e) {
                }
            });

            final JMenu plotScatter3D = new JMenu("Scatter plot (3D)");
            plotScatter3D.addMenuListener(new MenuListener() {
                @Override
                public void menuSelected(final MenuEvent e) {
                    if (book.getNumberOfSheets() == 0)
                        return;

                    final int sheetIndex = book.getActiveSheetIndex();
                    if (sheetIndex == -1)
                        return;

                    final Sheet sheet = book.getSheetAt(sheetIndex);
                    final IcySpreadSheet icySheet = Workbooks.getSheet(book, book.getSheetName(sheetIndex));

                    final JMenuItem labelFirstColumn = new JMenuItem("Select X axis...");
                    labelFirstColumn.setEnabled(false);
                    plotScatter3D.add(labelFirstColumn);
                    plotScatter3D.addSeparator();
                    for (int i = 0; i < icySheet.getNumberOfColumns(); i++) {
                        final double[] xValues = icySheet.getColumnValues(i);
                        if (xValues == null)
                            continue;

                        final double minX = ArrayMath.min(xValues);
                        final double maxX = ArrayMath.max(xValues);

                        final String xColumnName = tables.get(sheet).getColumnName(i);
                        final JMenu plotFirstColumn = new JMenu(xColumnName);
                        plotScatter3D.add(plotFirstColumn);

                        final JMenuItem labelSecondColumn = new JMenuItem("Select Y axis...");
                        labelSecondColumn.setEnabled(false);
                        plotFirstColumn.add(labelSecondColumn);
                        plotFirstColumn.addSeparator();

                        for (int j = 0; j < icySheet.getNumberOfColumns(); j++) {
                            if (i == j)
                                continue;

                            final double[] yValues = icySheet.getColumnValues(j);
                            if (yValues == null)
                                continue;

                            if (xValues.length != yValues.length)
                                throw new IcyHandledException(
                                        "Cannot create scatter plot: datasets have different sizes");

                            final double minY = ArrayMath.min(yValues);
                            final double maxY = ArrayMath.max(yValues);

                            final String yColumnName = tables.get(sheet).getColumnName(j);
                            final JMenu plotSecondColumn = new JMenu(yColumnName);
                            plotFirstColumn.add(plotSecondColumn);

                            final JMenuItem labelThirdColumn = new JMenuItem("Select Z axis...");
                            labelThirdColumn.setEnabled(false);
                            plotSecondColumn.add(labelThirdColumn);
                            plotSecondColumn.addSeparator();

                            for (int k = 0; k < icySheet.getNumberOfColumns(); k++) {
                                if (i == k || j == k)
                                    continue;

                                final double[] zValues = icySheet.getColumnValues(k);
                                if (zValues == null)
                                    continue;

                                if (xValues.length != zValues.length)
                                    throw new IcyHandledException(
                                            "Cannot create scatter plot: datasets have different sizes");

                                final double minZ = ArrayMath.min(zValues);
                                final double maxZ = ArrayMath.max(zValues);

                                final String zColumnName = tables.get(sheet).getColumnName(k);
                                final JMenuItem plotThirdColumn = new JMenuItem(zColumnName);
                                plotSecondColumn.add(plotThirdColumn);

                                plotThirdColumn.addActionListener(actionEvent -> {
                                    final Plot3DPanel panel = new Plot3DPanel();
                                    panel.addScatterPlot("Scatter plot", xValues, yValues, zValues);
                                    panel.setAxisLabels(xColumnName, yColumnName, zColumnName);
                                    panel.setFixedBounds(new double[]{minX, minY, minZ},
                                            new double[]{maxX, maxY, maxZ});

                                    final IcyFrame plot = new IcyFrame(
                                            xColumnName + " vs. " + yColumnName + " vs. " + zColumnName, true);
                                    plot.setPreferredSize(new Dimension(520, 420));
                                    plot.addToDesktopPane();
                                    plot.add(panel);
                                    plot.pack();
                                    plot.setVisible(true);
                                });
                            }
                        }
                    }
                }

                @Override
                public void menuDeselected(final MenuEvent e) {
                    plotScatter3D.removeAll();
                }

                @Override
                public void menuCanceled(final MenuEvent e) {

                }
            });

            plotPopupMenu.add(plotHistogram1D);
            plotPopupMenu.add(plotHistogram2D);
            plotPopupMenu.add(plotScatter2D);
            plotPopupMenu.add(plotScatter3D);
            plotButton.addActionListener(e -> plotPopupMenu.show(plotButton, 0, plotButton.getHeight()));
            optionBar.add(plotButton);

            // Plot (END)

            // Formula

            final VarEditor<String> formulaEditor = formula.createVarViewer();
            formulaEditor.setEnabled(true);
            optionBar.add((Component) formulaEditor.getEditorComponent());

            optionBar.add(Box.createHorizontalGlue());

            // Use header

            final JCheckBox useHeaderOption = (JCheckBox) useHeader.createVarEditor(true).getEditorComponent();
            useHeaderOption.setText("Use first row as header");
            useHeaderOption.setFocusable(false);
            optionBar.add(useHeaderOption);

            wbPanel.add(optionBar, BorderLayout.NORTH);

            if (tabs != null)
                tabs.removeChangeListener(tabChangeListener);
            tabs = new JTabbedPane(JTabbedPane.BOTTOM);
            tabs.addChangeListener(tabChangeListener);

            wbPanel.add(tabs, BorderLayout.CENTER);
            wbPanel.setName("" + book.hashCode());
        }

        final int nSheets = book.getNumberOfSheets();

        // remove unknown tabs (if any)
        tabCheck:
        for (int i = 0; i < tabs.getTabCount(); i++) {
            for (int j = 0; j < nSheets; j++)
                if (tabs.getTitleAt(i).equalsIgnoreCase(book.getSheetName(j)))
                    continue tabCheck;

            tabs.remove(i--);
        }

        for (int i = 0; i < nSheets; i++) {
            Sheet sheet = book.getSheetAt(i);

            // check if a tab exists for this sheet
            int currentSheetIndex = -1;

            for (int j = 0; j < tabs.getTabCount(); j++) {
                if (tabs.getTitleAt(j).equalsIgnoreCase(sheet.getSheetName())) {
                    currentSheetIndex = i;
                    break;
                }
            }

            if (currentSheetIndex == -1) {
                // create an empty tab to receive the table
                tabs.addTab(sheet.getSheetName(), new JPanel());
                currentSheetIndex = tabs.getTabCount() - 1;
            }

            sheet = book.getSheetAt(currentSheetIndex);

            if (tables.containsKey(sheet))
                updateTable(sheet);
            else
                createTable(sheet);
        }

        // remove old sheets from tables
        final List<Sheet> keys = new ArrayList<>(tables.keySet());
        for (final Sheet sheet : keys)
            if (book.getSheetIndex(sheet) == -1)
                tables.remove(sheet);

        // also remove from headers
        keys.clear();
        keys.addAll(headers.keySet());
        for (final Sheet sheet : keys)
            if (book.getSheetIndex(sheet) == -1)
                headers.remove(sheet);

        // Allow the creation of a new sheet
        if (!readOnly.getValue())
            tabs.addTab("+", new JPanel());
    }

    private static void createSheet(final @NotNull Workbook book) {
        book.createSheet("Sheet " + (book.getNumberOfSheets() + 1));
    }

    private void updateTable(final Sheet sheet) {
        final JXTable table = tables.get(sheet);
        final JXTable header = headers.get(sheet);

        final SheetModel model = (SheetModel) table.getModel();

        if (new IcySpreadSheet(sheet).getNumberOfColumns() != table.getColumnCount(true)) {
            model.fireTableStructureChanged();
        }
        else if (model.useFirstDataRowAsHeader != useHeader.getValue()) {
            model.useFirstDataRowAsHeader = useHeader.getValue();
            model.fireTableDataChanged();
            for (final TableColumn column : table.getColumns(true))
                column.setHeaderValue(table.getModel().getColumnName(column.getModelIndex()));
            table.getTableHeader().repaint();
        }

        // update the line numbers (on the left-hand side)
        header.tableChanged(new TableModelEvent(header.getModel()));
        table.tableChanged(new TableModelEvent(table.getModel()));
    }

    private void createTable(final Sheet sheet) {
        final IcySpreadSheet icySheet = new IcySpreadSheet(sheet);

        final int sheetIndex = icySheet.getIndex();

        final SheetModel tableModel = new SheetModel(icySheet, useHeader.getValue());
        final VarString sheetName = new VarString("sheet name", sheet.getSheetName());

        final JXTable table = new JXTable(tableModel) {
            @Contract("_, _ -> new")
            @Override
            public @NotNull TableCellEditor getCellEditor(final int row, final int column) {
                return new DefaultCellEditor(new JTextField()) {
                    @Override
                    public Component getTableCellEditorComponent(final JTable theTable, final Object value, final boolean isSelected, int theRow, int theColumn) {
                        theRow = convertRowIndexToModel(theRow);
                        theColumn = convertColumnIndexToModel(theColumn);

                        // formulas take the priority
                        String contents = icySheet.getFormula(theRow, theColumn);

                        contents = (contents.isEmpty() ? "" : "=") + icySheet.getValue(theRow, theColumn);

                        return super.getTableCellEditorComponent(theTable, contents, true, theRow, theColumn);
                    }
                };
            }

            @Contract("_, _ -> new")
            @Override
            public @NotNull TableCellRenderer getCellRenderer(final int row, final int column) {
                return new DefaultTableCellRenderer() {
                    @Override
                    public Component getTableCellRendererComponent(final JTable theTable, final Object value, final boolean isSelected, final boolean hasFocus, int theRow, int theColumn) {
                        theRow = convertRowIndexToModel(theRow);
                        theColumn = convertColumnIndexToModel(theColumn);

                        if (useHeader.getValue())
                            theRow++;

                        final Component component = super.getTableCellRendererComponent(theTable, value, isSelected, hasFocus, theRow, theColumn);

                        final Color fillColor = icySheet.getFillColor(theRow, theColumn);
                        if (fillColor != null)
                            component.setBackground(fillColor);

                        return component;
                    }
                };
            }
        };

        final Comparator<Object> comparator = (a, b) -> {
            if (a == null)
                a = "";
            if (b == null)
                b = "";

            // compare what is comparable
            if (a instanceof Comparable && b instanceof Comparable
                    && (a.getClass().isAssignableFrom(b.getClass()) || b.getClass().isAssignableFrom(a.getClass()))) {
                @SuppressWarnings("unchecked") final int compare = ((Comparable<Object>) a).compareTo(b);
                return compare;
            }

            // otherwise, one of them is perhaps a String
            if (a instanceof String)
                return -1;
            if (b instanceof String)
                return 1;

            // if none of the above, don't try comparing at all!
            return 0;
        };

        final int nCol = table.getColumnCount();
        for (int i = 0; i < nCol; i++) {
            table.getColumnExt(i).setComparator(comparator);
            table.setSortOrderCycle(SortOrder.DESCENDING, SortOrder.ASCENDING, SortOrder.UNSORTED);
        }

        table.setHorizontalScrollEnabled(true);
        table.setSelectionMode(ListSelectionModel.SINGLE_INTERVAL_SELECTION);
        table.setCellSelectionEnabled(true);
        table.setColumnControlVisible(true);
        table.setColumnControl(new PersistentColumnControlButton(table));

        final JScrollPane scrollPane = new JScrollPane(table, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);

        // Create the row header (first column with line numbers)
        // code adapted and largely inspired from:
        // http://stackoverflow.com/questions/6711877/jtable-use-row-numbers

        final AbstractTableModel headerModel = new AbstractTableModel() {
            @Contract(pure = true)
            @Override
            public int getColumnCount() {
                return 1;
            }

            @Contract(pure = true)
            @Override
            public @NotNull Object getValueAt(final int row, final int column) {
                return row + 1;
            }

            @Override
            public int getRowCount() {
                return tableModel.getRowCount();
            }
        };

        final JXTable headerTable = new JXTable(headerModel) {
            @Contract(value = "_, _ -> new", pure = true)
            @Override
            public @NotNull TableCellRenderer getCellRenderer(final int row, final int column) {
                return new TableCellRenderer() {
                    private final TableCellRenderer headerRenderer = table.getTableHeader().getDefaultRenderer();

                    Font headerFont = null;

                    @Override
                    public Component getTableCellRendererComponent(final JTable theTable, final Object value, boolean isSelected, final boolean hasFocus, final int theRow, final int theColumn) {
                        isSelected = table.getSelectionModel().isSelectedIndex(theRow);

                        final Component c = headerRenderer.getTableCellRendererComponent(theTable, value, isSelected,
                                hasFocus, -1, theColumn);

                        if (headerFont == null)
                            headerFont = c.getFont().deriveFont(isSelected ? Font.BOLD : Font.PLAIN);

                        c.setFont(headerFont);

                        return c;
                    }
                };
            }
        };

        headerTable.setShowGrid(true);
        headerTable.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
        headerTable.setPreferredScrollableViewportSize(new Dimension(40, 0));
        headerTable.getColumnModel().getColumn(0).setPreferredWidth(40);
        headerTable.setRowHeight(table.getRowHeight());

        // changes to the table should affect the header...

        table.getRowSorter().addRowSorterListener(e -> headerModel.fireTableDataChanged());

        // ... and clicking on the header should select the table line
        // headerTable.getSelectionModel().addListSelectionListener(new ListSelectionListener()
        // {
        // boolean eraseSelection = true;
        //
        // @Override
        // public void valueChanged(ListSelectionEvent selection)
        // {
        // if (eraseSelection) table.getSelectionModel().clearSelection();
        //
        // eraseSelection = false;
        //
        // int start = selection.getFirstIndex();
        // int end = selection.getLastIndex();
        //
        // // FIXME the end index always becomes the start index on the next click
        // table.setRowSelectionInterval(end, start);
        // table.setColumnSelectionInterval(0, table.getColumnCount() - 1);
        //
        // if (!selection.getValueIsAdjusting()) eraseSelection = true;
        // }
        // });

        scrollPane.setRowHeaderView(headerTable);

        // end of row header code

        tables.put(sheet, table);
        headers.put(sheet, headerTable);

        tabs.setComponentAt(sheetIndex, scrollPane);

        final JLabel sheetNameLabel = new JLabel(sheetName.getValue());

        class Listener implements ListSelectionListener, TableColumnModelListener {
            @Override
            public void valueChanged(final @NotNull ListSelectionEvent e) {
                if (e.getValueIsAdjusting())
                    return;

                if (table.getSelectedRow() == -1)
                    return;
                final int row = table.convertRowIndexToModel(table.getSelectedRow());
                final int column = table.convertColumnIndexToModel(table.getSelectedColumn());
                updateFormulaField(tableModel, row, column);
            }

            @Override
            public void columnSelectionChanged(final ListSelectionEvent e) {
                valueChanged(e);
            }

            @Contract(pure = true)
            @Override
            public void columnRemoved(final TableColumnModelEvent e) {
            }

            @Contract(pure = true)
            @Override
            public void columnMoved(final TableColumnModelEvent e) {
            }

            @Contract(pure = true)
            @Override
            public void columnMarginChanged(final ChangeEvent e) {
            }

            @Contract(pure = true)
            @Override
            public void columnAdded(final TableColumnModelEvent e) {
            }
        }

        final Listener l = new Listener();
        table.getSelectionModel().addListSelectionListener(l);
        table.getColumnModel().addColumnModelListener(l);

        if (!readOnly.getValue()) {
            // handle sheet name changing

            sheetNameLabel.setRequestFocusEnabled(false);

            sheetNameLabel.addMouseListener(new MouseAdapter() {
                @Override
                public void mouseClicked(final MouseEvent e) {
                    if (e.getClickCount() == 1) {
                        tabs.setSelectedIndex(sheetIndex);
                    }
                    else {
                        final Object answer = JOptionPane.showInputDialog(
                                sheetNameLabel,
                                null,
                                "Rename this sheet to...",
                                JOptionPane.QUESTION_MESSAGE,
                                null,
                                null,
                                sheetName.getValue()
                        );

                        if (answer != null)
                            sheetName.setValue(answer.toString());
                    }
                }
            });

            sheetName.addListener(new VarListener<>() {
                @Override
                public void valueChanged(final Var<String> source, final String oldValue, final String newValue) {
                    variable.getValue().setSheetName(sheetIndex, newValue);
                    sheetNameLabel.setText(newValue);
                }

                @Override
                public void referenceChanged(final Var<String> source, final Var<? extends String> oldReference, final Var<? extends String> newReference) {
                }
            });
        }

        tabs.setTabComponentAt(sheetIndex, sheetNameLabel);
        tabs.setTitleAt(sheetIndex, sheet.getSheetName());
    }

    private void updateFormulaField(final SheetModel model, int row, final int col) {
        if (col == -1) {
            formula.setValue("");
            return;
        }

        if (model.useFirstDataRowAsHeader)
            row++;

        String value = model.sheet.getFormula(row, col);

        if (!value.isEmpty()) {
            formula.setValue(" Formula: =" + value);
        }
        else {
            value = "" + model.sheet.getValue(row, col);

            if (!value.isEmpty()) {
                formula.setValue(" Value: " + value);
            }
            else {
                formula.setValue("");
            }
        }
    }

    private class SheetModel extends AbstractTableModel {
        final int MAX_NB_ROWS = 65536;

        final int MAX_NB_COLS = 256;

        private boolean useFirstDataRowAsHeader;

        final IcySpreadSheet sheet;

        public SheetModel(final IcySpreadSheet sheet, final boolean useFirstRowAsHeader) {
            this.sheet = sheet;
            this.useFirstDataRowAsHeader = useFirstRowAsHeader;
        }

        @Override
        public void setValueAt(final Object aValue, final int rowIndex, final int columnIndex) {
            if (aValue instanceof final String text) {
                if (text.isEmpty()) {
                    sheet.deleteCell(rowIndex, columnIndex);
                }
                else if (text.startsWith("="))
                    try {
                        // Parse a formula first
                        sheet.setFormula(rowIndex, columnIndex, text);
                    }
                    catch (final FormulaParseException e1) {
                        sheet.setValue(rowIndex, columnIndex, text);
                    }
                else
                    sheet.setValue(rowIndex, columnIndex, text);
            }
            else
                sheet.setValue(rowIndex, columnIndex, aValue);

            updateFormulaField(this, rowIndex, columnIndex);
        }

        @Override
        public Object getValueAt(int rowIndex, final int columnIndex) {
            if (useFirstDataRowAsHeader)
                rowIndex++;

            return sheet.getValue(rowIndex, columnIndex);
        }

        @Override
        public int getRowCount() {
            if (!readOnly.getValue())
                return MAX_NB_ROWS;

            final int nbRows = sheet.getNumberOfRows();

            return useFirstDataRowAsHeader ? nbRows - 1 : nbRows;
        }

        @Override
        public String getColumnName(final int columnIndex) {
            if (useFirstDataRowAsHeader)
                return "" + getValueAt(-1, columnIndex);

            return super.getColumnName(columnIndex);
        }

        @Override
        public int getColumnCount() {
            if (!readOnly.getValue())
                return MAX_NB_COLS;

            return sheet.getNumberOfColumns();
        }

        @Override
        public boolean isCellEditable(final int rowIndex, final int columnIndex) {
            return !readOnly.getValue();
        }
    }
}
