/*
 * Copyright (c) 2010-2024. Institut Pasteur.
 *
 * This file is part of Icy.
 * Icy is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Icy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Icy. If not, see <https://www.gnu.org/licenses/>.
 */

package plugins.adufour.vars.lang;

import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.ss.usermodel.WorkbookFactory;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.jetbrains.annotations.NotNull;
import plugins.adufour.vars.gui.VarEditor;
import plugins.adufour.vars.gui.swing.WorkbookEditor;
import plugins.adufour.vars.util.VarException;
import plugins.adufour.workbooks.Workbooks;

import javax.annotation.processing.FilerException;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

public class VarWorkbook extends Var<Workbook> {
    /**
     * Creates a new variable with the workbook read from the specified file
     *
     * @param name     the name of the workbook
     * @param workbook the workbook to work with
     */
    public VarWorkbook(final String name, final Workbook workbook) {
        super(name, Workbook.class, workbook, null);
    }

    /**
     * Creates a new variable with the workbook read from the specified file
     *
     * @param name the name of the workbook
     * @param file the file to read the workbook from
     * @throws IOException if input file is not found
     */
    public VarWorkbook(final String name, final File file) throws IOException {
        this(name, WorkbookFactory.create(file));
    }

    /**
     * Creates a new variable with a new empty workbook
     *
     * @param name           the name of the workbook
     * @param firstSheetName the name of the first sheet
     */
    public VarWorkbook(final String name, final String firstSheetName) {
        this(name, Workbooks.createEmptyWorkbook());

        getValue().createSheet(firstSheetName);
    }

    @Override
    public VarEditor<Workbook> createVarEditor() {
        return new WorkbookEditor(this);
    }

    @Override
    public VarEditor<Workbook> createVarViewer() {
        final WorkbookEditor editor = new WorkbookEditor(this);
        editor.setReadOnly(true);
        editor.setOpenButtonVisible(false);
        return editor;
    }

    /**
     * Save the workbook to disk
     *
     * @param folder       the folder on disk where the workbook should be saved (must exist)
     * @param workbookName the name of the workbook (without extension)
     * @throws VarException if the workbook is <code>null</code>
     * @throws IOException  if the file cannot be accessed
     */
    public void saveToDisk(final @NotNull File folder, final String workbookName) throws VarException, IOException {
        if (!folder.isDirectory())
            throw new FilerException(folder + "is not a valid folder");

        final Workbook wb = getValue(true);

        String filename = folder.getPath() + File.separator + workbookName + ".xls";

        if (wb instanceof XSSFWorkbook)
            filename = filename + "x";

        final FileOutputStream out = new FileOutputStream(filename);
        wb.write(out);
        out.close();
    }
}
