/*
 * Copyright (c) 2010-2024. Institut Pasteur.
 *
 * This file is part of Icy.
 * Icy is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Icy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Icy. If not, see <https://www.gnu.org/licenses/>.
 */

package plugins.adufour.workbooks;

import org.apache.poi.hssf.usermodel.HSSFPalette;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.hssf.util.HSSFColor;
import org.apache.poi.ss.formula.FormulaParseException;
import org.apache.poi.ss.usermodel.*;
import org.apache.poi.xssf.usermodel.DefaultIndexedColorMap;
import org.apache.poi.xssf.usermodel.XSSFCellStyle;
import org.apache.poi.xssf.usermodel.XSSFColor;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.apache.xmlbeans.impl.values.XmlValueDisconnectedException;
import org.bioimageanalysis.icy.system.logging.IcyLogger;
import org.jetbrains.annotations.NotNull;

import java.awt.Color;
import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

/**
 * Results table for Icy. This class relies on the Apache POI library, and acts as a high-level
 * wrapper for the {@link Sheet} class
 *
 * @author Alexandre Dufour
 */
public class IcySpreadSheet {
    private final Sheet sheet;

    private final FormulaEvaluator evaluator;

    private final Map<Color, CellStyle> colorStyleMap;

    public IcySpreadSheet(final @NotNull Sheet sheet) {
        this.sheet = sheet;
        this.evaluator = sheet.getWorkbook().getCreationHelper().createFormulaEvaluator();
        this.colorStyleMap = new HashMap<>();
    }

    /**
     * @param sourceColumn the index of the row to copy (NB: the first row is at index 0)
     * @param targetColumn the index of the row where the source row will be pasted (NB: the first row is at
     *                     index 0)
     */
    public void copyColumn(final int sourceColumn, final int targetColumn) {
        final Row source = sheet.getRow(sourceColumn);
        final Row target = sheet.getRow(targetColumn);

        if (source == null) {
            if (target != null)
                sheet.removeRow(target);
            return;
        }

        if (target == null)
            sheet.createRow(targetColumn);

        final int nCells = source.getLastCellNum();

        for (int i = 0; i < nCells; i++) {
            final Cell sourceCell = source.getCell(i, Row.MissingCellPolicy.RETURN_BLANK_AS_NULL);

            if (sourceCell != null) {
                assert target != null;
                final Cell targetCell = target.getCell(i, Row.MissingCellPolicy.CREATE_NULL_AS_BLANK);
                Workbooks.copyCell(sourceCell, targetCell);
            }
        }
    }

    /**
     * @param sourceRow the index of the row to copy (NB: the first row is at index 0)
     * @param targetRow the index of the row where the source row will be pasted (NB: the first row is at
     *                  index 0)
     */
    public void copyRow(final int sourceRow, final int targetRow) {
        final Row source = sheet.getRow(sourceRow);
        final Row target = getOrCreateRow(targetRow);

        if (source == null) {
            if (target != null)
                sheet.removeRow(target);
            return;
        }

        if (target == null)
            sheet.createRow(targetRow);

        final int nCells = source.getLastCellNum();

        for (int i = 0; i < nCells; i++) {
            final Cell sourceCell = source.getCell(i, Row.MissingCellPolicy.RETURN_BLANK_AS_NULL);

            if (sourceCell != null) {
                assert target != null;
                final Cell targetCell = target.getCell(i, Row.MissingCellPolicy.CREATE_NULL_AS_BLANK);
                Workbooks.copyCell(sourceCell, targetCell);
            }
        }
    }

    /**
     * Deletes the cell at the specified row and column (NB: the first row or column index is 0).
     *
     * @param rowIndex    the row index of the cell to delete (starting at 0)
     * @param columnIndex the column index of the cell to delete (starting at 0)
     */
    public void deleteCell(final int rowIndex, final int columnIndex) {
        final Row row = sheet.getRow(rowIndex);
        if (row == null)
            return;

        final Cell cell = row.getCell(columnIndex);
        if (cell != null) {
            row.removeCell(cell);
            try {
                evaluator.notifyDeleteCell(cell);
            }
            catch (final XmlValueDisconnectedException e) {
                IcyLogger.warn(this.getClass(), e); // Should not appear
            }
        }
    }

    /**
     * Deletes the column at the specified index (NB: the first column is at index 0).
     *
     * @param columnIndex the index of the column to remove (starting at 0)
     * @param shiftData   set to <code>true</code> to shift all other columns up, or <code>false</code> to
     *                    leave the column empty
     */
    public void deleteColumn(final int columnIndex, final boolean shiftData) {
        final int lastRow = sheet.getLastRowNum();

        for (int r = 0; r <= lastRow; r++) {
            final Row row = sheet.getRow(r);
            if (row == null)
                continue;

            final int lastColumn = row.getLastCellNum() - 1;

            if (columnIndex > lastColumn)
                continue;

            final Cell cell = row.getCell(columnIndex);

            if (cell != null)
                row.removeCell(cell);

            if (shiftData && columnIndex < lastColumn) {
                for (int i = columnIndex; i < lastColumn; i++)
                    Workbooks.copyCell(row.getCell(i + 1), row.getCell(i, Row.MissingCellPolicy.CREATE_NULL_AS_BLANK));
                row.removeCell(row.getCell(lastColumn));
            }
        }
    }

    /**
     * Deletes the row at the specified index (NB: the first row is at index 0).
     *
     * @param rowIndex  the index of the row to remove (starting at 0)
     * @param shiftData set to <code>true</code> to shift all other lines up, or <code>false</code> to
     *                  leave the line empty
     */
    public void deleteRow(final int rowIndex, final boolean shiftData) {
        final int lastRow = sheet.getLastRowNum();

        // don't do anything if the column does not exist
        if (rowIndex >= lastRow)
            return;

        final Row rowToRemove = sheet.getRow(rowIndex);

        if (rowToRemove != null)
            sheet.removeRow(rowToRemove);

        if (shiftData && rowIndex < lastRow) {
            for (int i = rowIndex + 1; i <= lastRow; i++)
                copyRow(i + 1, i);

            sheet.removeRow(sheet.getRow(lastRow));
        }
    }

    /**
     * @param row    row index
     * @param column column index
     * @return the background color of the specified cell
     */
    public Color getFillColor(final int row, final int column) {
        final Cell cell = getOrCreateRow(row).getCell(column, Row.MissingCellPolicy.CREATE_NULL_AS_BLANK);

        final org.apache.poi.ss.usermodel.Color color = cell.getCellStyle().getFillForegroundColorColor();
        if (color instanceof HSSFColor) {
            final short[] rgb = ((HSSFColor) color).getTriplet();
            if ((rgb != null) && (rgb.length >= 3)) {
                if (rgb[0] != 0 || rgb[1] != 0 || rgb[2] != 0) {
                    return new Color(rgb[0], rgb[1], rgb[2]);
                }
            }
        }
        else if (color instanceof XSSFColor) {
            final byte[] rgb = ((XSSFColor) color).getRGB();
            if ((rgb != null) && (rgb.length >= 3)) {
                if (rgb[0] != 0 || rgb[1] != 0 || rgb[2] != 0) {
                    return new Color(rgb[0] & 0xff, rgb[1] & 0xff, rgb[2] & 0xff);
                }
            }
        }

        return null;
    }

    /**
     * @param rowIndex    the 0-based row index of the cell
     * @param columnIndex the 0-based column index of the cell
     * @return the cell formula at the specified location (or an empty string if the cell does not
     * hold a formula)
     */
    public String getFormula(final int rowIndex, final int columnIndex) {
        final Row row = sheet.getRow(rowIndex);
        if (row == null)
            return "";

        final Cell cell = row.getCell(columnIndex, Row.MissingCellPolicy.RETURN_BLANK_AS_NULL);
        if (cell == null)
            return "";

        return cell.getCellType() == CellType.FORMULA ? cell.getCellFormula() : "";
    }

    /**
     * @param rowIndex    the 0-based row index of the cell
     * @param columnIndex the 0-based column index of the cell
     * @return the cell at the specified location (or an empty string if the cell is empty)
     */
    public Object getValue(final int rowIndex, final int columnIndex) {
        final Row row = sheet.getRow(rowIndex);
        if (row == null)
            return "";

        final Cell cell = row.getCell(columnIndex, Row.MissingCellPolicy.RETURN_BLANK_AS_NULL);
        if (cell == null)
            return "";

        if (cell.getCellType() == CellType.STRING)
            return cell.getStringCellValue();

        if (cell.getCellType() == CellType.BOOLEAN)
            return cell.getBooleanCellValue();

        return evaluator.evaluate(cell).getNumberValue();
    }

    /**
     * @param rowIndex the 0-based row index
     * @return
     */
    private Row getOrCreateRow(final int rowIndex) {
        final Row row = sheet.getRow(rowIndex);
        return row != null ? row : sheet.createRow(rowIndex);
    }

    /**
     * @return the index of this sheet within its workbook (NB: the first index is 0)
     */
    public int getIndex() {
        return sheet.getWorkbook().getSheetIndex(sheet);
    }

    /**
     * @return the name of this sheet
     */
    public String getName() {
        return sheet.getSheetName();
    }

    /**
     * @return The number of rows in this table. Note that this number indicates the index of the
     * last row, but does not necessarily indicate that all other rows contain data
     */
    public int getNumberOfRows() {
        return sheet.getLastRowNum() + 1;
    }

    /**
     * @return The number of columns in this table. Note that this number indicates the index of the
     * last known column, but does not necessarily indicate that all other columns contain
     * data, or that all rows use that many columns
     */
    public int getNumberOfColumns() {
        int nCols = 0;
        final int lastRow = sheet.getLastRowNum();
        for (int i = 0; i <= lastRow; i++) {
            final Row row = sheet.getRow(i);
            if (row != null)
                nCols = Math.max(nCols, row.getLastCellNum());
        }
        return nCols;
    }

    /**
     * @return the underlying {@link Sheet} implementation backing this object.
     */
    public Sheet getSheet() {
        return sheet;
    }

    /**
     * Sets the fill color of the specified cell
     *
     * @param row    row index
     * @param column column index
     * @param color  the color to assign
     */
    public void setFillColor(final int row, final int column, final Color color) {
        final Cell cell = getOrCreateRow(row).getCell(column, Row.MissingCellPolicy.CREATE_NULL_AS_BLANK);

        final Workbook book = sheet.getWorkbook();
        // try to find cell style if we already created it
        CellStyle style = colorStyleMap.get(color);

        try {
            // not found ?
            if (style == null) {
                // create it
                style = book.createCellStyle();

                // apply color
                if (book instanceof HSSFWorkbook) {
                    final HSSFPalette palette = ((HSSFWorkbook) book).getCustomPalette();
                    final short colorIndex = palette.findSimilarColor(color.getRed(), color.getGreen(), color.getBlue()).getIndex();
                    style.setFillForegroundColor(colorIndex);
                }
                else if (book instanceof XSSFWorkbook) {
                    ((XSSFCellStyle) style).setFillForegroundColor(new XSSFColor(color, new DefaultIndexedColorMap()));
                }

                style.setFillPattern(FillPatternType.SOLID_FOREGROUND);

                // add style to map
                colorStyleMap.put(color, style);
            }

            // set style to cell
            cell.setCellStyle(style);
        }
        catch (final Exception e) {
            // just ignore any other possible error as
            // java.lang.IllegalStateException: The maximum number of Cell Styles was exceeded.
        }
    }

    /**
     * Sets the formula for the specified cell. The formula follows the standard Excel-compatible
     * syntax with alphanumeric references to other cells (e.g. "=A1*A2"). <br>
     * NOTE: this method is not responsible for checking the formula for potential syntax errors.
     *
     * @param row     row index
     * @param column  column index
     * @param formula the formula
     * @throws FormulaParseException if the formula has errors
     */
    public void setFormula(final int row, final int column, String formula) throws FormulaParseException {
        final Cell cell = getOrCreateRow(row).getCell(column, Row.MissingCellPolicy.CREATE_NULL_AS_BLANK);

        if (formula.startsWith("="))
            formula = formula.substring(1);

        cell.setCellFormula(formula);

        evaluator.notifySetFormula(cell);
    }

    /**
     * Sets the cell value at the specified location. Values of the following types are supported:
     * <ul>
     * <li>{@link String} (plain text)</li>
     * <li>{@link Boolean} (<code>true</code> or <code>false</code>)</li>
     * <li>Any compatible {@link Number} (will be converted to double precision)</li>
     * </ul>
     * Other unsupported objects will be replaced by their String representation (via
     * {@link Object#toString()})<br>
     * NOTE: to assign a formula to this cell, use {@link #setFormula(int, int, String)}.
     *
     * @param row    the index of the row containing the cell (NB: the first row is at index 0)
     * @param column the index of the column containing the cell (NB: the first column is at index 0)
     * @param value  the value to assign to the cell. NB: the type of cell will be automatically
     *               guessed from the provided value, and if no standard format is found (number,
     *               boolean, text), its String representation will be used
     */
    public void setValue(final int row, final int column, Object value) {
        if (value == null) {
            deleteCell(row, column);
            return;
        }

        final Cell cell = getOrCreateRow(row).getCell(column, Row.MissingCellPolicy.CREATE_NULL_AS_BLANK);

        // what is the value like?

        if (value instanceof final String text) {
            try {
                value = Double.parseDouble(text);
            }
            catch (final NumberFormatException numberE) {
                // Parse boolean values manually (because Boolean.parseBoolean() never fails)
                if (text.equalsIgnoreCase("true")) {
                    value = true;
                }
                else if (text.equalsIgnoreCase("false")) {
                    value = false;
                }
            }
        }

        // a number?
        if (Number.class.isAssignableFrom(value.getClass())) {
            cell.setCellValue(((Number) value).doubleValue());
        }
        else if (value instanceof Boolean) {
            cell.setCellValue((Boolean) value);
        }
        // if not, default to a String representation
        else {
            cell.setCellValue(value.toString());
        }

        evaluator.notifyUpdateCell(cell);
    }

    /**
     * Sets an entire column of values from a list of values
     *
     * @param columnIndex  the index of the target column (Note: the first column is at index 0)
     * @param columnValues an array of values to add, line by line. NB: numbers and string will be recognized
     *                     as such, any other item will be converted to string using
     *                     {@link Object#toString()}, while <code>null</code> values will result in an empty
     *                     cell
     */
    public void setColumn(final int columnIndex, final Object... columnValues) {
        setColumn(columnIndex, null, columnValues);
    }

    /**
     * Sets an entire column of values from a list of values
     *
     * @param columnHeader a header text placed in an extra row before the data (or <code>null</code> if not
     *                     needed)
     * @param columnIndex  the index of the target column (Note: the first column is at index 0)
     * @param columnValues an array of values to add, line by line. NB: numbers and string will be recognized
     *                     as such, any other item will be converted to string using
     *                     {@link Object#toString()}, while <code>null</code> values will result in an empty
     *                     cell
     */
    public void setColumn(final int columnIndex, final String columnHeader, final Object... columnValues) {
        int rowIndex = 0;
        if (columnHeader != null) {
            final Row currentRow = getOrCreateRow(rowIndex);
            currentRow.getCell(columnIndex, Row.MissingCellPolicy.CREATE_NULL_AS_BLANK).setCellValue(columnHeader);
            rowIndex = 1;
        }

        for (final Object obj : columnValues) {
            if (obj == null) {
                rowIndex++;
                continue;
            }

            setValue(rowIndex++, columnIndex, obj);
        }
    }

    /**
     * Insert a new empty column at the specified index, shifting all columns to the right
     *
     * @param columnIndex the index where the new column is to be inserted (NB: the first column index is 0)
     */
    public void insertColumn(final int columnIndex) {
        // As painfully as it might (not) sound, there is no built-in function for this!
        final Iterable<Row> iterable = sheet::rowIterator;

        for (final Row row : iterable)
            for (int i = row.getPhysicalNumberOfCells() - 1; i >= columnIndex; i--) {
                final Cell oldCell = row.getCell(i);
                final Cell newCell = row.createCell(i + 1, oldCell.getCellType());
                Workbooks.copyCell(oldCell, newCell);
            }
    }

    /**
     * Convenience method that sets an entire row of values at the specified location in a given
     * workbook
     *
     * @param rowIndex the index of the row to fill (Note: the first column is at index 0)
     * @param values   the values to add, separated by commas, or a single array containing the values to
     *                 add.<br>
     *                 NB: numbers and string will be recognized as such, any other item will be
     *                 converted to string using {@link Object#toString()}, while <code>null</code>
     *                 values will result in an empty cell
     */
    public void setRow(final int rowIndex, final Object @NotNull ... values) {
        final Iterable<?> list;

        // special case: rowValues could contain a single array
        if (values.length == 1 && values[0].getClass().isArray()) {
            final Object array = values[0];
            final int n = Array.getLength(array);
            final ArrayList<Object> arrayList = new ArrayList<>(n);
            for (int i = 0; i < n; i++)
                arrayList.add(Array.get(array, i));
            list = arrayList;
        }
        else {
            list = Arrays.asList(values);
        }

        setRow(rowIndex, null, list);
    }

    /**
     * Sets an entire row of values at the specified (0-based) row index
     *
     * @param header    a header text placed in the first column, before the data (or <code>null</code> if
     *                  not needed)
     * @param rowIndex  the index of the target row (Note: the first row is at index 0)
     * @param rowValues an array of values to add, column by column. NB: numbers and string will be
     *                  recognized as such, any other item will be converted to string using
     *                  {@link Object#toString()}, while <code>null</code> values will result in an empty
     *                  cell
     */
    public void setRow(final int rowIndex, final String header, final Iterable<?> rowValues) {
        int colIndex = 0;
        if (header != null)
            setValue(rowIndex, colIndex++, header);

        for (final Object obj : rowValues)
            if (obj != null)
                setValue(rowIndex, colIndex++, obj);
    }

    /**
     * Removes all rows in this sheet
     */
    public void removeRows() {
        removeRows(0);
    }

    /**
     * Removes all rows in this sheet starting at the specified (0-based) row index
     *
     * @param startRowIndex the (0-based) index of the first row to remove
     */
    public void removeRows(final int startRowIndex) {
        for (int i = startRowIndex; i <= sheet.getLastRowNum(); i++) {
            final Row row = sheet.getRow(i);
            if (row != null)
                sheet.removeRow(row);
        }
    }

    /**
     * @param columnIndex the column to search
     * @return All values in the column corresponding to a number (empty cells and other cell
     * formats are discarded)
     */
    public double[] getColumnValues(final int columnIndex) {
        final ArrayList<Double> values = new ArrayList<>(sheet.getLastRowNum() + 1);
        for (int i = 0; i <= sheet.getLastRowNum(); i++) {
            final Object val = getValue(i, columnIndex);
            if (val instanceof Number)
                values.add(((Number) val).doubleValue());
        }
        if (values.isEmpty())
            return null;

        final double[] result = new double[values.size()];
        for (int i = 0; i < result.length; i++)
            result[i] = values.get(i);
        return result;
    }
}
